package org.implab.gradle.mercurial;

import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.gradle.api.Project;

import groovy.lang.Closure;

public class MercurialExtension {
    private static final String COMMIT_INFO_TEMPLATE = "{latesttag('re:^v') % '{ifeq(tag,'null','',tag)}:{distance}'}:{branch}:{node|short}";

    private static final Pattern versionPattern = Pattern.compile("^(?:v[\\-_\\.]?)?(.*)");

    private final Project project;

    private Charset outputCharset = Charset.defaultCharset();

    private WorkspaceInfo workspaceInfo;

    private String hgCmd = "hg";

    private String stagingBookmark = "staging";

    private String releaseBookmark = "prod";

    private SemVersion defaultWorkspaceVersion = new SemVersion(0, 0, 1, "dev", null);

    private Function<SemVersion, SemVersion> preReleasePolicy;

    public MercurialExtension(Project project) {
        Objects.requireNonNull(project);

        this.project = project;
    }

    public String hg(String... args) {
        List<String> commandLine = new ArrayList<>();
        commandLine.add(hgCmd);
        Collections.addAll(commandLine, args);

        return Utils.execCmd(commandLine, outputCharset.name());
    }

    public String getReleaseBookmark() {
        return releaseBookmark;
    }

    public void setReleaseBookmark(String value) {
        if(value == null)
            throw new IllegalArgumentException();
        releaseBookmark = value;
    }

    public String getStagingBookmark() {
        return stagingBookmark;
    }

    public void setStagingBookmark(String value) {
        if(value == null)
            throw new IllegalArgumentException();
        stagingBookmark = value;
    }

    public Charset getOutputCharset() {
        return outputCharset;
    }

    public void setOutputCharset(Charset outputCharset) {
        this.outputCharset = Optional.ofNullable(outputCharset).orElseGet(Charset::defaultCharset);
    }

    public void setOutputCharset(String outputCharset) {
        this.outputCharset = Optional.ofNullable(outputCharset).map(x -> Charset.forName(x))
                .orElseGet(Charset::defaultCharset);
    }

    public String getHgCmd() {
        return hgCmd;
    }

    public void setHgCmd(String hgCmd) {
        this.hgCmd = hgCmd;
    }

    public WorkspaceInfo getWorkspaceInfo() {
        if (workspaceInfo == null)
            workspaceInfo = loadWorkspaceInfo();
        return workspaceInfo;
    }

    public SemVersion getWorkspaceVersion() {
        return tryParseVersion(getWorkspaceInfo().getVersionTag()).orElse(defaultWorkspaceVersion);
    }

    private SemVersion defaultPreReleasePolicy(SemVersion version) {
        String changeset = getWorkspaceInfo().getChangeset();
        return version.addPatch(1).withPreRelease("snapshot").withMeta(changeset);
    }

    public void preReleasePolicy(Closure<SemVersion> policy) {
        if (policy == null)
            throw new IllegalArgumentException();

        this.preReleasePolicy = x -> {
            policy.setDelegate(getWorkspaceInfo());
            policy.setResolveStrategy(Closure.DELEGATE_FIRST);
            return policy.call(x);
        };
    }

    public void preReleasePolicy(Function<SemVersion, SemVersion> policy) {
        if (policy == null)
            throw new IllegalArgumentException();
        this.preReleasePolicy = policy;
    }

    public void applyVersioningPolicy() {
        Object version = project.getVersion();
        if (version == null || version.toString().isBlank()
                || version.toString().equalsIgnoreCase(Project.DEFAULT_VERSION)) {
            // if the version isn't specified
            int distance = getWorkspaceInfo().getVersionDistance();

            // the current workspace is not tagged with the version
            if (distance > 0) {
                SemVersion preReleaseVersion = Optional.ofNullable(preReleasePolicy)
                        .orElse(this::defaultPreReleasePolicy)
                        .apply(getWorkspaceVersion());

                project.setVersion(preReleaseVersion);
            } else {
                project.setVersion(getWorkspaceVersion());
            }
        } else {
            project.setVersion(SemVersion.toSemVersion(version, false));
        }
    }

    String[] queryLogInfo() {
        return hg("log", "-r", ".", "--template", COMMIT_INFO_TEMPLATE).split(":");
    }

    WorkspaceInfo loadWorkspaceInfo() {
        boolean isDirty = checkIsDirty();
        String[] info = queryLogInfo();
        return new WorkspaceInfo(info[0], Integer.parseInt(info[1]), info[2], info[3], isDirty);
    }

    boolean checkIsDirty() {
        return !hg("status", "-mard").isBlank();
    }

    Optional<SemVersion> tryParseVersion(String version) {
        if (version == null)
            return Optional.empty();

        Matcher m = versionPattern.matcher(version);
        boolean isMatch = m.matches();

        return isMatch ? SemVersion.tryParseLax(m.group(1)) : Optional.empty();
    }

}
